#AUTOGENERATED! DO NOT EDIT! File to edit: dev/losses.ipynb (unless otherwise specified).

__all__ = ['mse', 'mae', 'q_loss', 'tilted_loss', 'weighted_loss']

#Cell
import torch
import torch.nn as nn
import torch.nn.functional as F

#Cell
def mse(y_pred, y, weights=None):
    """
    Mean Squared Error
    """
    return weighted_loss((y_pred - y)**2, weights)

def mae(y_pred, y, weights=None):
    """
    Mean Absolute Error
    """
    return weighted_loss(torch.abs(y_pred - y), weights)

def q_loss(y_pred, y, quantiles = [0.05, 0.5, 0.95], weights=None):
    """
    y_pred: Predicted Value
    y: Target
    quantiles: Quantile
    weights(optional): Weighting of prediction-target pair.
    """
    if isinstance(quantiles, list):
        e = 0
        for q, pred in zip(quantiles, y_pred.split(1, dim=-1)):
            e += weighted_loss(tilted_loss(pred, y, q), weights)
        return e.mean()
    else:
        return weighted_loss(tilted_loss(y_pred, y, quantiles), weights)

def tilted_loss(y_pred, y, q=0.5):
    """
    y_pred: Predicted Value
    y: Target
    q: Quantile
    """
    e = (y - y_pred)
    return q * torch.clamp_min(e, 0) + (1-q) * torch.clamp_min(-e, 0)

def weighted_loss(loss, weights):
    """
    Weighted loss
    """
    if weights is not None:
        return torch.mean(weights * loss)
    else:
        return torch.mean(loss)


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/callbacks.ipynb (unless otherwise specified).

__all__ = ['L1Loss', 'PrintLoss']

#Cell
import torch
from fastai.basics import LearnerCallback

#Cell
class L1Loss(LearnerCallback):
    def __init__(self, learn, beta=1e-2):
        super().__init__(learn)
        self.beta = beta

    def on_backward_end(self, **kwargss):
        weights = {k:v for k,v in self.learn.model.named_parameters()
                   if not 'bias' in k}

        for k, v in weights.items():
            sign = torch.ones_like(v)
            sign[v<0] = -1
            v = v - self.beta * sign
            self.learn.model.state_dict()[k].copy_(v)

#Cell
def _get_loss(model, dl, loss_func):
    """Calculate `loss_func` of `model` on `dl` in evaluation mode."""
    model.eval()
    with torch.no_grad():
        loss = 0
        N = 0
        for xb,yb in dl:
            # breakpoint()
            if isinstance(xb, list):
                loss += loss_func(model(*xb), yb) * len(yb)
            else:
                loss += loss_func(model(xb), yb) * len(yb)
            N += len(yb)
        return loss / N

class PrintLoss(LearnerCallback):
    """
    Prints Loss in one line as opposed to fastai Printer
    """
    def __init__(self, learn):
        super().__init__(learn)

    def on_epoch_end(self, **kwargs):
        train_loss = _get_loss(self.learn.model, self.learn.data.train_dl, loss_func=self.learn.loss_func)
        val_loss = _get_loss(self.learn.model, self.learn.data.valid_dl, loss_func=self.learn.loss_func)

        epoch = kwargs['epoch']
        n_epochs = kwargs['n_epochs']

        print(f'Epoch {epoch+1}/{n_epochs} Training Loss: {train_loss:.4f}, Validation Loss: {val_loss:.4f}', end='\r')


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/data.ipynb (unless otherwise specified).

__all__ = ['tt', 'convert_date', 'get_moments', 'create_tensors', 'TimeSeries', 'DataFrame', 'create_db']

#Cell
import torch
from torch.utils.data import Dataset, DataLoader
from fastai.data_block import DataBunch, DatasetType

import warnings

torch.Tensor.ndim = property(lambda x: x.dim())
tt = torch.Tensor

#Cell
_MILISECONDS_IN_DAY = 1e9*3600*24
def convert_date(dates):
    dates = dates.astype(int) / _MILISECONDS_IN_DAY
    return torch.Tensor(dates).reshape(-1,1)

def get_moments(df):
    """
    Gets the mean and variance of all columns.
    """
    df.ds = df.ds.astype(int) / _MILISECONDS_IN_DAY
    mean = df.mean()
    std = df.std()
    moments = {
        't': [mean['ds'], std['ds']],
        't_range': [df.ds.min(), df.ds.max()],
        'y': [mean['y'], std['y']]
        }
    mean.drop(['ds', 'y'], inplace=True)
    std.drop(['ds', 'y'], inplace=True)

    for k, v in moments.items():
        if v[1] == 0:
            warnings.warn(f'Standard deviation of {k} is 0. Resetting to 1.')
            moments[k][1] = 1.0

    if len(mean) > 0: # there are x variables
        moments['x'] = [tt(mean.values[None,:]),
                        tt(std.values[None,:])]
        if (moments['x'][1] == 0).any():
            idx = moments['x'][1] == 0
            moments['x'][1][idx] = 1
            warnings.warn('Standard deviation of some x feature is 0. Resetting to 1.')

    return moments

def create_tensors(df, moments, predict=False):
    """
    converts a pandas dataframe to pytorch tensors
    """
    # get time tensor
    t = convert_date(df['ds'].values)
    data = {'t': t}
    df.drop(['ds'], axis=1, inplace=True)

    # get y tensor (if not in predict stage)
    if not predict: # 'y' in df.columns and
        y = torch.Tensor(df['y'].values).reshape(-1, 1)
        df.drop(['y'], axis=1, inplace=True)
        data['y'] = y

    if predict and 'y' in df.columns:
        df.drop(['y'], axis=1, inplace=True)

    # add x if it's available
    if df.shape[1] > 0:
        x = torch.Tensor(df.values).float()
        data['x'] = x


    data = {k: (v - moments[k][0]) / moments[k][1] for k, v in data.items()}

    return data, moments

#Cell
class TimeSeries(Dataset):
    def __init__(self, x, y):
        super().__init__()
        self.x, self.y = x, y

    def __len__(self):
        return len(self.y)

    def __getitem__(self, i):
        return torch.Tensor([self.x[i]]), torch.Tensor([self.y[i]])

class DataFrame(Dataset):
    def __init__(self, df, moments=None):
        super().__init__()
        self.data, self.moments = create_tensors(df, moments)

    def __len__(self):
        return len(self.data['t'])

    def __getitem__(self, i):
        if 'x' in self.data:
            return (self.data['t'][i], self.data['x'][i]), self.data['y'][i]
        else:
            return self.data['t'][i], self.data['y'][i]

#Cell
def create_db(df, train_p=0.8, bs=96, moments=None):
    train_len = int(train_p*len(df))
    df.reset_index(drop=True, inplace=True)
    train_ds = DataFrame(df.iloc[:train_len], moments)
    val_ds = DataFrame(df.iloc[train_len:], moments)

    bs = min(bs, len(train_ds))
    val_bs = min(bs, len(val_ds))
    return DataBunch.create(train_ds, val_ds, bs=bs, val_bs=val_bs)


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/model.ipynb (unless otherwise specified).

__all__ = ['Model']

#Cell
import numpy as np
import matplotlib.pyplot as plt
from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F

from fastai.basics import Learner, LearnerCallback, Callback

from .blocks import DefaultModel, DefaultQModel
from ..data.data import create_db, create_tensors, get_moments
from ..losses import q_loss
from ..callbacks.callbacks import *

#Cell
class Model(nn.Module):
    """
    Default Model Wrapper
    parameters:
    - df: dataset used in training dataset.
    - model (optional): how to model time series. Default: DefaultModel.
    - loss (optional): loss function: Default l1 loss.
    - bs (optional): batchsize
    - alpha (optional): l2 weight decay
    - beta (optional): l1 weight decay
    - silent: whether to silence output or not

    Usage:
    ```
    model = Model(train_df)
    model.fit(train_df)
    y = model.predict(test_df)
    ```
    """
    def __init__(self, df, model=None, model_args=None, quantiles=[0.05, 0.5, 0.95],
                 loss=None, bs=128, epochs=100, lr=0.3, alpha=0, beta=0.1, silent=True):
        super().__init__()
        self.moments = get_moments(df.copy())

        if loss is None:
            self.model = DefaultQModel(self.moments, **model_args, quantiles=quantiles)
            self.loss = partial(q_loss, quantiles=quantiles)
        else:
            self.model = DefaultModel(self.moments, **model_args)
            self.loss = loss
        self.lr = lr
        self.alpha = alpha
        self.beta = beta
        self.bs = bs
        self.epochs = epochs
        self.silent = silent

    def fit(self, df):
        # self.find_appropriate_lr(df)
        learner = self.create_learner(df)
        cb = [L1Loss(learner, self.beta), PrintLoss(learner)]
        # breakpoint()

        learner.fit_one_cycle(self.epochs, self.lr, callbacks=cb)

        self.model = learner.model

    def predict(self, df):
        x, _ = create_tensors(df.copy(), self.moments, predict=True)
        mean, sd = self.moments['y']
        y = sd * self.model(**x) + mean
        return y.detach().cpu().numpy()

    def forward(self, *args):
        return self.model(*args)

    def create_learner(self, df):
        db = create_db(df, bs=self.bs, moments=self.moments)
        # breakpoint()
        learner = Learner(db, self.model, loss_func=self.loss, wd=self.alpha, silent=True)
        return learner


#AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Trend": "blocks.ipynb",
         "LinearX": "blocks.ipynb",
         "FourierModel": "blocks.ipynb",
         "Seasonal": "blocks.ipynb",
         "Holiday": "blocks.ipynb",
         "HolidayRange": "blocks.ipynb",
         "Squasher": "blocks.ipynb",
         "RandomWalk": "blocks.ipynb",
         "DefaultModel": "blocks.ipynb",
         "DefaultQModel": "blocks.ipynb",
         "L1Loss": "callbacks.ipynb",
         "PrintLoss": "callbacks.ipynb",
         "torch.Tensor.ndim": "data.ipynb",
         "tt": "data.ipynb",
         "convert_date": "data.ipynb",
         "get_moments": "data.ipynb",
         "create_tensors": "data.ipynb",
         "TimeSeries": "data.ipynb",
         "DataFrame": "data.ipynb",
         "create_db": "data.ipynb",
         "mse": "losses.ipynb",
         "mae": "losses.ipynb",
         "q_loss": "losses.ipynb",
         "tilted_loss": "losses.ipynb",
         "weighted_loss": "losses.ipynb",
         "Model": "model.ipynb"}

modules = ["model/blocks.py",
           "callbacks/callbacks.py",
           "data/data.py",
           "losses.py",
           "model/model.py"]

git_url = "https://github.com/fastai/profetorch/tree/master/"

def custom_doc_links(name): return None


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/blocks.ipynb (unless otherwise specified).

__all__ = ['Trend', 'LinearX', 'FourierModel', 'Seasonal', 'Holiday', 'HolidayRange', 'Squasher', 'RandomWalk',
           'DefaultModel', 'DefaultQModel']

#Cell
import numpy as np
import matplotlib.pyplot as plt

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..data.data import convert_date

#Cell
class Trend(nn.Module):
    """
    Broken Trend model, with breakpoints as defined by user.
    """
    def __init__(self, breakpoints:int=None, moment=None):
        super().__init__()
        self.init_layer = nn.Linear(1,1) # first linear bit

        if breakpoints is not None:
            if isinstance(breakpoints, int):
                # range = moment['t_range'][1] - moment['t_range'][0]
                # breakpoints = torch.rand(breakpoint)*range + moment['t_range'][0]
                if breakpoints > 0:
                    breakpoints = np.linspace(*moment['t_range'], breakpoints+1, endpoint=False)[1:]
                else:
                    breakpoints = None
            else:
                breakpoints = convert_date(breakpoints)
            # create deltas which is how the gradient will change
            deltas = torch.zeros(len(breakpoints)) # initialisation
            self.deltas = nn.Parameter(deltas) # make it a parameter

        self.bpoints = breakpoints

    def __copy2array(self):
        """
        Saves parameters into wb
        """
        # extract gradient and bias
        w = self.init_layer.weight
        b = self.init_layer.bias
        self.params = [[w,b]] # save it to buffer
        if self.bpoints:
            for d, x1 in zip(self.deltas, self.bpoints):
                y1 = w *x1 + b # find the endpoint of line segment (x1, y1)
                w = w + d # add on the delta to gradient
                b = y1 - w * x1 # find new bias of line segment
                self.params.append([w,b]) # add to buffer

        # create buffer
        self.wb = torch.zeros(len(self.params), len(self.params[0]))
        for i in range(self.wb.shape[0]):
            for j in range(self.wb.shape[1]):
                self.wb[i,j] = self.params[i][j]

    def forward(self, t:torch.Tensor):
        if self.bpoints is not None:
            self.__copy2array() # copy across parameters into matrix
            # get the line segment area (x_sec) for each x
            x_sec = t >= self.bpoints
            x_sec = x_sec.sum(1)

            # get final prediction y = mx +b for relevant section
            return t*self.wb[x_sec][:,:1] + self.wb[x_sec][:,1:]

        else:
            return self.init_layer(t)

#Cell
class LinearX(nn.Module):
    """
    Linear model of non-time based inputs.
    """
    def __init__(self, dims):
        super().__init__()
        if dims > 0:
            self.linear = nn.Linear(dims, 1, bias=False)

    def forward(self, x):
        if x is not None:
            return self.linear(x)
        else:
            return 0

#Cell
_TWOPI = 2*np.pi

class FourierModel(nn.Module):
    """
    Block that outputs sin's and cos' as basis functions.
    """
    def __init__(self, p:float=365.25, scale:float=1, n:int=7):
        super().__init__()
        self.np = [(i+1, p/scale) for i in range(n)]
        if n > 0:
            self.linear = nn.Linear(n * 2, 1, bias=False)

    def forward(self, t:torch.Tensor):
        if len(self.np) > 0:
            cos = [torch.cos(_TWOPI * n * t / p) for n,p in self.np]
            sin = [torch.sin(_TWOPI * n * t / p) for n,p in self.np]

            return self.linear(torch.cat(cos + sin, dim=1))

        else:
            return 0

    def plot(self):
        if self.n > 0:
            t = torch.linspace(0, self.p, steps=100)
            y = self.forward(t[:,None])
            plt.figure(figsize=(12,5))
            plt.plot(t.cpu().numpy(), y.detach().cpu().numpy())
            plt.show()


class Seasonal(nn.Module):
    """
    Combination of Fourier Blocks that gives weekly, monthly and
    yearly seasonality.
    """
    def __init__(self, y_n=7, m_n=5, w_n=0,
                 y_p=365.25, m_p=30.5, w_p=7, scale=1):
        super().__init__()
        self.yearly = FourierModel(y_p, scale, y_n) # , w[:,:idxs[0]]
        self.monthly = FourierModel(m_p, scale, m_n) # w[:,idxs[0]:idxs[1]]
        self.weekly = FourierModel(w_p, scale, w_n) # w[:,idxs[1]:idxs[2]]

    def forward(self, t):
        return self.yearly(t) + self.monthly(t) + self.weekly(t)

#Cell
class Holiday(nn.Module):
    def __init__(self, holiday, repeat_every=365, mean=0, scale=1):
        super().__init__()
        self.holiday = (holiday - mean) / scale
        self.repeat_every = repeat_every / scale
        self.w = nn.Parameter(torch.zeros(1)+0.05)

    def forward(self, t):
        rem = torch.remainder(t - self.holiday, self.repeat_every)
        return (rem == 0).float() * self.w


class HolidayRange(nn.Module):
    def __init__(self, holidays):
        """
        holidays: list of lists containing lower and upper bound of hols
        """
        super().__init__()
        self.holidays = holidays
        self.w = nn.Parameter(torch.zeros(1)+0.05)

    def forward(self, t):
        bounded = [(l<=t) & (t<=u) for l,u in self.holidays]
        return sum(bounded).float()*self.w

#Cell
class Squasher(nn.Module):
    """
    Squashes output to lie beween `high` and `low`.
    """
    def __init__(self, low=None, high=None, mean=0, sd=1, alpha=0.01):
        super().__init__()
        if low is not None:
            low = (low - mean) / sd
        if high is not None:
            high = (high - mean) / sd
        self.L, self.H, self.alpha = low, high, alpha

    def forward(self, t):
        if self.L is not None:
            t[t < self.L] = self.alpha * (t[t < self.L] - self.L) + self.L
        if self.H is not None:
            t[t > self.H] = self.alpha * (t[t > self.H] - self.H) + self.H
        return t


class RandomWalk(nn.Module):
    def __init__(self, n, breaks):
        super().__init__()
        self.w = nn.Parameter(torch.randn(n,1))
        self.breaks = breaks

    def forward(self, x):
        w = F.softplus(self.w.cumsum(0))
        x_sec = get_section(x, self.breaks)
        return w[x_sec]

#Cell
class DefaultModel(nn.Module):
    """
    Sum of Linear Trend, Seasonality and squashed.
    """
    def __init__(self, moments, breakpoints=None, y_n=7, m_n=5, w_n=0, l=None, h=None):
        super().__init__()
        if 'x' in moments:
            dims = moments['x'][0].shape[1]
        else:
            dims = 0

        self.trend = Trend(breakpoints, moments)
        self.seasonal = Seasonal(y_n, m_n, w_n, scale=moments['t'][1])
        self.linear = LinearX(dims)
        self.squash = Squasher(l, h, *moments['y'])

    def forward(self, t, x=None):
        prediction = self.seasonal(t) + self.trend(t) + self.linear(x)
        prediction = self.squash(prediction)
        return prediction

class DefaultQModel(nn.Module):
    """
    Same as DefaultModel but with multiple outputs corresponding to quantiles
    """
    def __init__(self, moments, breakpoints=None, y_n=7, m_n=5, w_n=0, l=None, h=None, quantiles=[0.05, 0.5, 0.95]):
        super().__init__()
        assert 0.5 in quantiles, f'0.5 needs to be in quantiles. Provided {quantiles} as quantiles.'
        self.idx = quantiles.index(0.5)
        signs = [q-0.5 for q in quantiles]
        self.signs = torch.Tensor([-1 if s<0 else 1 for i,s in enumerate(signs) if i != self.idx])[None,:]
        self.idxs = [i for i in range(len(quantiles)) if i != self.idx]
        median_args = {'y_n': y_n, 'm_n': m_n, 'w_n': w_n}
        other_args = {'y_n': 0, 'm_n': 0, 'w_n': 0}
        args = [median_args if q==0.5 else other_args for q in quantiles]
        self.models = nn.ModuleList([DefaultModel(moments, breakpoints, **arg) for arg in args])
        self.squash = Squasher(l, h, *moments['y'])

    def forward(self, t, x=None):
        prediction = torch.cat([m(t,x) for m in self.models], -1)
        median = prediction[:, [self.idx]]
        prediction[:, self.idxs] = median + F.softplus(prediction[:, self.idxs]) * self.signs

        return self.squash(prediction)